@@ -1,5 +1,4 @@
 import logging
-import math
 import time
 from dataclasses import dataclass
 from itertools import repeat
@@ -8,8 +7,10 @@ from typing import List, Union
 
 import numpy as np
 import pandas as pd
+import alphashape
 from scipy.spatial import Delaunay
 from scipy.special import gamma
+from scipy.spatial.qhull import QhullError
 from shapely import geometry
 from shapely.ops import cascaded_union, polygonize, triangulate, unary_union
 from sklearn.cluster import DBSCAN
@@ -28,7 +29,6 @@ class FootprintOutput:
     purity: float
 
 
-# _empty_footprint = FootprintOutput(geometry.MultiPolygon(), 0, 0, 0, 0, 0)
 def _empty_footprint() -> FootprintOutput:
     return FootprintOutput(geometry.MultiPolygon(), 0, 0, 0, 0, 0)
 
@@ -86,6 +86,17 @@ def alpha_shape(points: np.ndarray, alpha=1.0):
     return cascaded_union(triangles), edge_points
 
 
+def alphashape_optm(points: np.ndarray, alpha_ini):
+    points = list(map(tuple, points))
+    hull = alphashape.alphashape(points, alpha_ini)
+    if isinstance(hull, geometry.Polygon):
+        return hull
+    else:
+        alpha = 0.9 * alphashape.optimizealpha(points)
+        hull = alphashape.alphashape(points, alpha)
+        return hull
+
+
 def count_points(pts):
     if pts.is_empty:
         return 0
@@ -182,7 +193,7 @@ def trace(Z: np.ndarray, Ybin: np.ndarray, P: np.ndarray, beta: np.ndarray, algo
         _logger.debug(f"Base algorithm {repr(algolabels[i])} completed.")
 
     _logger.info("TRACE is calculating the beta-footprint.")
-    hard = trace_build(Z, ~beta, kwargs['PI'])
+    hard = trace_build_wrapper(Z, ~beta, kwargs['PI'])
 
     _logger.info("TRACE is preparing the summary table.")
     summary = make_summary(space, good, best, algolabels)
@@ -197,7 +208,7 @@ def trace_build_wrapper(*args):
         return FootprintOutput(geometry.MultiPolygon(), 0, 0, 0, 0, 0)
 
 
-def trace_build(Z, Ybin, PI):
+def trace_build(Z, Ybin, PI) -> FootprintOutput:
     Ig = np.unique(Z[Ybin, :], axis=0)
     if Ig.shape[0] < 3:
         raise TraceException
@@ -211,8 +222,16 @@ def trace_build(Z, Ybin, PI):
         polydata = Ig[clustering.labels_ == i, :]
         if polydata.shape[0] < 3:
             continue
-        concave_hull, edge_points = alpha_shape(polydata, alpha=0.5)
-        boundary = np.array(concave_hull.exterior.xy).T
+        # concave_hull, edge_points = alpha_shape(polydata, alpha=0.5)
+        try:
+            concave_hull = alphashape_optm(polydata, 0.5)
+        except QhullError:
+            continue
+
+        if concave_hull.is_empty:
+            raise TraceException
+        else:
+            boundary = np.array(concave_hull.exterior.xy).T
         aux = trace_fitpoly(boundary, Z, Ybin, PI)
         if not aux.is_empty:
             polygon = polygon.union(aux)
@@ -315,8 +334,16 @@ def trace_tight(polygon: Union[geometry.Polygon, geometry.MultiPolygon],
         polydata = np.array(criteria)
         if polydata.shape[0] < 3:
             continue
-        concave_hull, edge_points = alpha_shape(polydata, alpha=0.5)
-        boundary = np.array(concave_hull.exterior.xy).T
+        # concave_hull, edge_points = alpha_shape(polydata, alpha=0.5)
+        try:
+            concave_hull = alphashape_optm(polydata, 0.5)
+        except QhullError:
+            continue
+
+        if concave_hull.is_empty:
+            continue
+        else:
+            boundary = np.array(concave_hull.exterior.xy).T
         aux = trace_fitpoly(boundary, Z, Ybin, *args)
         if aux.is_empty:
             continue
